#!/usr/bin/env python
#
#
#   Copyright (c) 2007 Jesse Guardiani <jesse@guardiani.us>
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License as
#   published by the Free Software Foundation; either version 2 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
#   02111-1307, USA.
#

import pygame, time
from pygame.locals import *
import globals
from db      import db      as db
from theme   import theme   as theme
from manager import manager as manager


class BaseWidget(pygame.sprite.Sprite):
  def __init__(self):
    pygame.sprite.Sprite.__init__(self) # call Sprite initializer
    self.image = None
    self.rect  = None

  def handle_event(self,event):
    ''' override me '''

  def is_idle(self):
    ''' override me '''
    return 1

  def update(self):
    ''' override me '''

  def draw(self,surface):
    surface.blit(self.image,self.rect)

  def close(self):
    ''' override me '''


class ScrollWidget(BaseWidget):
  ''' This widget wraps sprites with equal heights and 
  presents the user with an interface for scrolling and
  selecting these sprites'''
  def __init__(self,rect):
    BaseWidget.__init__(self) # call Group initializer
    self.rect      = pygame.Rect(rect) # this is where we draw our widget within the surface
    self.y_spacer  = 5 # height in pixels of whitespace between the top and bottom of sprites
    self.point_x   = self.rect.left # this tracks the absolute position of the upper left corner of self.get_sprites()[0]
    self.point_y   = self.rect.top
    self.distance  = 0
    self.mousedown = 0
    self.speed     = 0
    self.num_events= 0
    self.friction  = 2
    self.friction_c= 0
    self.sprites        = []
    self.sorted_sprites = []
    self.upper_bar = Bar(Rect(self.rect.left,0,self.rect.width,5))
    self.lower_bar = Bar(Rect(self.rect.left,0,self.rect.width,5))
    self._get_selection_rect()
    self.selected_index = None
    self.need_sort = False

  def _get_selection_rect(self):
    ''' The selection rect is used to determine which sprite is currently selected'''
    left   = self.rect.left
    top    = self.upper_bar.rect.top
    width  = self.rect.width
    height = self.lower_bar.rect.bottom - self.upper_bar.rect.top
    self.selection_rect = Rect(left,top,width,height)
  
  def _sort_sprites(self):
    if (self.need_sort): # cache sort order between list updates
      print "sorting"
      self.need_sort = False
      self.sorted_sprites = self.sprites
      self.sorted_sprites.sort()

  def _calc_selected_sprite(self):
    ''' This method uses the selection_rect to determine which sprite is selected.'''
    if self.speed or self.mousedown:
      if self.has_selected(): self.get_selected_sprite().set_focus(False)
      self.selected_index = None
      return # no selected_sprite when scrolling

    if self.has_selected(): return

    for sprite in self.sorted_sprites: # FIXME: use viewport point to skip to the first possible sprite
      # determine selected sprite
      if self.selection_rect.collidepoint(sprite.rect.center):
        self.set_selected_sprite(sprite)
        print "Selected"
        return

    print "none found!" # this should never happen

  def _snap_to_selection(self):
    ''' center selected sprite in view'''
    if self.has_selected():
      snap_y = self.selection_rect.top - self.get_selected_sprite().rect.top + self.y_spacer
      self.point_y = self.point_y + snap_y
      print "point_y=%s,selection_rect.top=%s,get_selected_sprite().rect.top=%s" % (self.point_y,self.selection_rect.top,self.get_selected_sprite().rect.top)

  def add(self,*sprite):
    self.need_sort = True
    self.sprites.append(*sprite)

  def remove(self,*sprite):
    self.need_sort = True
    self.sprites.remove(*sprite)

  def has_selected(self):
    if self.selected_index == None: return False
    return True

  def next(self):
    ''' Move selected_sprite to next sprite in list '''
    if self.is_idle() and self.has_selected():
      self.set_selected_index(self.selected_index + 1)

  def prev(self):
    ''' Move selected_sprite to previous sprite in list '''
    if self.is_idle() and self.has_selected():
      self.set_selected_index(self.selected_index - 1)

  def get_selected_sprite(self):
    if not self.has_selected(): return None
    return self.sorted_sprites[self.selected_index]

  def set_selected_index(self,index):
    if self.has_selected(): self.get_selected_sprite().set_focus(False)
    if index < 0: index = len(self.sorted_sprites) - 1
    elif index > len(self.sorted_sprites) - 1: index = 0
    self.selected_index = index
    print "selected=%s" % (self.selected_index,)
    self.get_selected_sprite().set_focus(True)
    self.update()
    self._snap_to_selection()

  def set_selected_sprite(self,new_selected_sprite):
    print "select it"
    c = 0
    for sprite in self.sorted_sprites: # make sure this new selected sprite exists in our list
      if new_selected_sprite == sprite:
        self.set_selected_index(c)
        return True
      c += 1
    print "invalid set_selected!" # this should never happen
    return False

  def get_sprites(self):
    self._sort_sprites()
    return self.sorted_sprites

  def update(self):
    if self.speed: # apply fake friction to reduce speed
        self.point_y = self.point_y + self.speed
        self.friction_c = self.friction_c + 1
        if self.friction_c >= self.friction:
          self.friction_c = 0
          self.speed = self.speed - cmp(self.speed,0)

    self._calc_selected_sprite()
    if not self.speed and not self.has_selected() and not self.mousedown:
      print "speed is 0 and no selected_sprite"

    self._sort_sprites()
    sprite_l = self.sorted_sprites
    if not sprite_l: return

    # some necessary calculations
    head_sprite     = sprite_l[0]
    viewport_h      = len(sprite_l) * (head_sprite.rect.height + self.y_spacer)
    half_viewport_h = viewport_h / 2
    self.upper_bar.rect.top = self.rect.centery - head_sprite.rect.height / 2 - self.y_spacer
    self.lower_bar.rect.top = self.rect.centery + head_sprite.rect.height / 2
    self._get_selection_rect()
   
    i = 0
    for sprite in sprite_l:
      # calculate position of each sprite
      sprite.rect.left = self.point_x
      sprite.rect.top  = self.point_y + (i * (head_sprite.rect.height + self.y_spacer))
      if sprite.rect.top > self.rect.centery + half_viewport_h: # wrap bottom
        sprite.rect.top = sprite.rect.top - viewport_h
        if i == 0: self.point_y = sprite.rect.top
      if sprite.rect.top < self.rect.centery - half_viewport_h: # wrap top
        sprite.rect.top = sprite.rect.top + viewport_h
        if i == 0: self.point_y = sprite.rect.top
      i = i + 1

    if self.has_selected(): self.get_selected_sprite().update()

  def handle_event(self,event):
    if event.type == MOUSEBUTTONDOWN:
      if self.rect.collidepoint(pygame.mouse.get_pos()):
        manager.has_focus = True
        self.speed      = 0
        self.distance   = 0
        self.num_events = 0
        self.mousedown = 1
    elif event.type == MOUSEBUTTONUP:
      if self.mousedown:
        self.mousedown  = 0
        self.friction_c = 0
        if self.num_events <= 12 and abs(self.distance) <= 12:
          self.speed      = 0
          self.distance   = 0
          self.num_events = 0
    elif event.type == MOUSEMOTION:
      if self.mousedown and self.sorted_sprites:
        (rel_x,rel_y) = event.rel
        self.point_y  = self.point_y + rel_y
        if cmp(rel_y,0) and cmp(rel_y,0) != cmp(self.distance,0): # changed direction
          self.num_events = 0
          self.distance   = 0
        self.num_events = self.num_events + 1
        self.distance   = self.distance   + rel_y
        self.speed      = self.sorted_sprites[0].rect.height / 50 * self.distance / self.num_events

  def is_idle(self):
    return (not self.speed and not self.mousedown)

  def draw(self,surface):
    for widget in self.sorted_sprites:
      if self.rect.colliderect(widget.rect): widget.draw(surface)
    surface.blit(self.upper_bar.image,self.upper_bar.rect)
    surface.blit(self.lower_bar.image,self.lower_bar.rect)

  def stop_animation(self):
    self.speed      = 0
    self.distance   = 0
    self.num_events = 0
    self.mousedown  = 0
    self.friction_c = 0
    self.update()

  def close(self):
    for widget in self.sorted_sprites: widget.remove()


class Album(BaseWidget):
  def __init__(self,name):
    BaseWidget.__init__(self)
    self.name       = name
    self.theme_rect = theme.get_rect('album_art_bg')
    self.cache_rect = None
    self.rect       = None
    self.get_cache_rect()
    self.rect       = Rect(0,0,self.theme_rect.width,self.theme_rect.height)

  def get_cache_rect(self):
    list = db.get_album_art(self.name)
    if not list:
      print "invalid album cache for %s" % (self.name,)
      return # should never happen

    row = list[0]
    self.cache_rect = Rect(row['x'],row['y'],self.theme_rect.width,self.theme_rect.height)

  def draw(self,surface,point=None):
    if not point: point = self.rect
    surface.blit(globals.albumcache,point,self.cache_rect)

  def __cmp__(self,other):
    return cmp(self.name.lower(),other.name.lower())


class Button(BaseWidget):
#
# we could also add a is_escape=False for albumbuttons etc below, which would then binds to the esc key
# but is it the right approach? -disq
#
  def __init__(self,image,onclick_cb,is_default=False):
    BaseWidget.__init__(self)
    self.image = image
    self.rect  = self.image.get_rect()
    self.onclick_cb = onclick_cb
    self.mousedown = False
    self.mouseup_timeout_start = None
    self.is_default = is_default

  def handle_event(self,event):
    if event.type == MOUSEBUTTONDOWN:
      # only register a button click if there has been some time since last button click. This prevents false clicks.
      if not self.mouseup_timeout_start or time.time()-self.mouseup_timeout_start > 0:
        self.mousedown = True
        self.mouseup_timeout_start = None
    if self.mousedown and event.type == MOUSEBUTTONUP:
      if self.rect.collidepoint(pygame.mouse.get_pos()):
        self.mousedown = False
        self.mouseup_timeout_start = time.time()
        self.onclick_cb()
    if self.is_default and event.type == KEYDOWN and event.key == K_RETURN:
        self.onclick_cb()


class PlayBackButton(Button):
  ''' like a Button, but only visible when playlist is not empty '''

  def handle_event(self,event):
    if not manager.playlist.list: return
    Button.handle_event(self,event)

  def draw(self,surface):
    if not manager.playlist.list: return
    Button.draw(self,surface)


class PlayButton(PlayBackButton):
  def __init__(self):
    ''' I really hope I don't have to explain this button. '''
    self.image = theme.get_surface((125,43))
    theme.blit(self.image,(50,0),'play')
    PlayBackButton.__init__(self,self.image,self._make_callback())
    self.rect.left = 655 # FIXME: the controlling view should probably set the x,y pos for us
    self.rect.top  = 160
    self.playing   = False

  def _load_image(self):
    if manager.playlist.is_paused(): theme.blit(self.image,(50,0),'play')
    else: theme.blit(self.image,(50,0),'pause')

  def update(self):
    self._load_image()

  def _make_callback(self):
    def _callback():
      manager.playlist.pause()
      self.update()
      return 
    return _callback


class NextButton(PlayBackButton):
  def __init__(self):
    ''' Skips forward by one track '''
    self.image = theme.get_surface((125,43))
    theme.blit(self.image,(50,0),'next')
    PlayBackButton.__init__(self,self.image,self._make_callback())
    self.rect.left = 655 # FIXME: the controlling view should probably set the x,y pos for us
    self.rect.top  = 225

  def _make_callback(self):
    def _callback():
      manager.playlist.next(None, None)
      return 
    return _callback


class PrevButton(PlayBackButton):
  def __init__(self):
    ''' Skips backward by one track '''
    self.image = theme.get_surface((125,43))
    theme.blit(self.image,(50,0),'prev')
    PlayBackButton.__init__(self,self.image,self._make_callback())
    self.rect.left = 655 # FIXME: the controlling view should probably set the x,y pos for us
    self.rect.top  = 100

  def _make_callback(self):
    def _callback():
      manager.playlist.prev(None, None)
      return 
    return _callback


class A2DPButton(Button):
  def __init__(self):
    ''' This is the bluetooth button in the upper right. It displays
    the status of bluez's a2dpd. '''
    self.image = theme.get_surface((125,43))
    Button.__init__(self,self.image,self._make_a2dp_button_callback())
    if not manager.has_a2dp: return
    theme.blit(self.image,(45,0),'a2dp_off')
    self.rect.left = 655
    self.rect.top  = 10

  def _load_image(self):
    import mplayer
    if isinstance(manager.player, mplayer.Mplayer):
      a2dp = manager.player.get_a2dpd()
    else: a2dp = False
    print "a2dp: %s" % (a2dp,)
    self.image.fill(theme.colorkey)
    if a2dp: theme.blit(self.image,(45,0),'a2dp_on')
    else: theme.blit(self.image,(45,0),'a2dp_off')

  def update(self):
    if not manager.has_a2dp: return
    self._load_image()

  def _make_a2dp_button_callback(self):
    def _a2dp_button_callback():
      print "switching a2dp modes from button"
      manager.clear()
      status_area = Text("PLEASE WAIT...")
      status_area.rect.topleft = (200,200)
      status_area.update()
      status_area.draw(manager.screen)
      pygame.display.flip()
      import mplayer
      if not isinstance(manager.player, mplayer.Mplayer):
        # switch to mplayer for a2dp
        target = manager.player.target
        paused = manager.player.paused
        volume = manager.player.volume
        manager.player.close()
        manager.player = mplayer.Mplayer()
        if target:
          manager.player.play(target)
          manager.player.set_volume(volume,show_volumeview=False)
        if target and paused: manager.player.pause(True)
      manager.player.set_a2dpd(not manager.player.get_a2dpd())
      if not manager.player.get_a2dpd():
        # switch to normal player for non a2dp
        target = manager.player.target
        paused = manager.player.paused
        volume = manager.player.volume
        manager.player.close()
        manager._get_player()
        if target:
          manager.player.play(target)
          manager.player.set_volume(volume,show_volumeview=False)
        if target and paused: manager.player.pause(True)
      self.update()
      return 
    return _a2dp_button_callback


class NowPlayingButton(Button):
  def __init__(self):
    ''' This is the button in the lower right. It displays
    the currently playing album and song. '''
    self.title_text_x = 2
    self.title_text_y = 2
    self.position_text_x = 2
    self.position_text_y = 20
    self.album_x = 0
    self.album_y = 40
    self.image = theme.get_surface((125,160+self.album_y))
    self.pbar_background = None
    theme.blit(self.image,(self.album_x,self.album_y),'album_art_bg')
    Button.__init__(self,self.image,self._make_nowplaying_button_callback())
    self.rect.left = 655 # FIXME: the controlling view should probably set the x,y pos for us
    self.rect.top  = manager.screen.get_rect().height - self.image.get_rect().height
    self.current   = None
    self.last_album_sprite = None
    self.position_string = ""
    self.current_song = None
    self.pbar         = None
    self._get_pbar_background()

  def _get_pbar_background(self):
    self.pbar_background = theme.get_surface((self.rect.width,self.album_y - self.position_text_y))
    self.pbar_background.blit(manager.background,(0,0),Rect(
      self.rect.left,
      self.rect.top  + self.position_text_y,
      self.rect.width,
      self.album_y - self.position_text_y))
    self.pbar_background.fill((0,0,0),Rect(0,self.pbar_background.get_rect().height - 5,self.pbar_background.get_rect().width,5))

  def _load_image(self):
    # position text
    if self.current_song and manager.playlist.status() == self.current:
      song = self.current_song
    else:
      song = self.current_song = db.song_of_path(manager.playlist.status())[0]
    if not self.pbar: self.pbar = ProgressBar(
        Rect(0,self.album_y - 4,self.rect.width,3),segmented=False)

    current_time   = "%2d:%02d" % divmod(int(manager.player.seconds),60)
    #id3_length     = "%2d:%02d" % divmod(int(song['length']         ),60)
    player_length = "%2d:%02d" % divmod(int(manager.player.length ),60)
    #new_pos_s = "%s of %s/%s (%3d%%)" % (current_time,id3_length,player_length,manager.player.percent)
    new_pos_s = "%s of %s (%3d%%)" % (current_time,player_length,manager.player.percent)
    if self.position_string != new_pos_s:
      self.pbar.set_percent(manager.player.percent)
      self.position_string = new_pos_s
      self.image.blit(self.pbar_background,(0,self.position_text_y))
      self.pbar.blit(self.image)
      globals.gprint(
        self.position_string,
        self.image,
        (self.position_text_x, self.position_text_y),
        16,
        False)
    if manager.playlist.status() == self.current: return
    print "nowplaying: " + song['album']
    if not self.last_album_sprite or self.last_album_sprite.name != song['album']:
      for sprite in manager.album_view.album_widget.get_sprites():
        if sprite.album.name == song['album']:
          self.last_album_sprite = sprite.album
          break
    self.last_album_sprite.draw(self.image,(self.album_x,self.album_y))

    track = song['track']
    if not track: track = ''
    else: track = "%02d " % (track,)
    # title text
    self.image.blit(manager.background,(self.title_text_x,self.title_text_y),Rect(
      self.rect.left + self.title_text_x,
      self.rect.top  + self.title_text_y,
      self.rect.width - self.title_text_x,
      self.position_text_y - self.title_text_y))
    globals.gprint(
      track+song['title'],
      self.image,
      (self.title_text_x, self.title_text_y),
      16,
      False)
    self.current = manager.playlist.status()

  def update(self):
    if not manager.playlist.status():
      return
    self._load_image()

  def _make_nowplaying_button_callback(self):
    def _nowplaying_button_callback():
      print "show now_playing"
      return manager.show_nowplaying()
    return _nowplaying_button_callback



class AlbumArtist(BaseWidget):
  def __init__(self,album,artist):
    BaseWidget.__init__(self)
    self.album    = Album(album)
    self.artist   = Text(artist,album)
    width         = self.album.rect.width + self.artist.rect.width
    height        = self.album.rect.height
    self.rect     = Rect(0,0,width,height)

  def draw(self,surface):
    self.album.rect.topleft = (self.rect.left,self.rect.top)
    self.album.draw(surface)
    self.artist.rect.topleft = (
        self.rect.left + 125,
        self.rect.top + (self.album.rect.height / 2 - self.artist.rect.height / 2))
    self.artist.draw(surface)

  def update(self):
    self.album.update()
    self.artist.update()

  def set_focus(self,focus):
    self.artist.set_focus(focus)
      
  def __cmp__(self,other):
    return cmp(
        ': '.join((self.artist.name.lower(),self.album.name.lower())),
        ': '.join((other.artist.name.lower(),other.album.name.lower()))
    )


# This is **not** a widget. It's more of a font wrapper. It makes it easy to implement scrolling text within a rectangle.
class ScrollingText():
  def __init__(self,name,rect,size=False):
    self.rect     = rect
    self.name     = name
    self.size     = size
    self.offset   = 0
    self.step     = 10  # number of pixels to move text
    self.frequency= 0.1 # how often to move text, in seconds
    self.focus    = False
    self.font_surf= globals.gprint_ret(self.name,self.size)
    self.sep_font_surf= globals.gprint_ret(' ~ ',self.size)
    self.scroll_timeout_start = 0

  def draw(self,surface):
    old_clip = surface.get_clip()
    surface.set_clip(self.rect)
    surface.blit(self.font_surf,(self.rect.left-self.offset,self.rect.top))
    if self.offset:
      surface.blit(self.sep_font_surf,(self.rect.left-self.offset + self.font_surf.get_rect().width,self.rect.top))
      surface.blit(self.font_surf,(self.rect.left-self.offset + self.font_surf.get_rect().width + self.sep_font_surf.get_rect().width,self.rect.top))
    surface.set_clip(old_clip)
      
  def set_focus(self,focus):
    print "set_focus: %s for %s" % (focus,self.name)
    if focus != self.focus:
      if not focus:
        self.offset = 0
    self.focus = focus

  def update(self):
    if not self.scroll_timeout_start or time.time() - self.scroll_timeout_start >= self.frequency:
      self.scroll_timeout_start = time.time()
      if self.font_surf.get_rect().width> self.rect.width:
        self.offset = self.offset + self.step
        if self.offset >= self.font_surf.get_rect().width + self.sep_font_surf.get_rect().width: self.offset = 0
        return True
    return False


class Text(BaseWidget):
  def __init__(self, name, subtext=None, selected=False):
    BaseWidget.__init__(self)
    self.name     = name
    self.subtext  = subtext
    self.selected = selected
    self.theme_rect = None
    self.get_theme_rect()
    self.rect     = Rect(0,0,self.theme_rect.width,self.theme_rect.height)
    self.focus    = False
    self.scrolling_text = None
    self.scrolling_subtext = None
    self.text_x   = 5
    self.text_y   = 33
    if self.subtext: self.text_y = self.text_y - 20
    self.get_scrolling_text()
    self.get_scrolling_subtext()

  def get_theme_rect(self):
    if self.selected:
      self.theme_rect = theme.get_rect('text_bg_sel')
    else:
      self.theme_rect = theme.get_rect('text_bg')

  def get_scrolling_text(self):
    if not self.scrolling_text:
      self.scrolling_text = ScrollingText(
          self.name,
          Rect(self.text_x, self.text_y, self.rect.width - (self.text_x * 2), self.rect.height))

  def get_scrolling_subtext(self):
    if self.subtext and not self.scrolling_subtext:
      self.scrolling_subtext = ScrollingText(
          self.subtext,
          Rect(self.text_x, self.text_y+30, self.rect.width - (self.text_x * 2), self.rect.height),
          30)

  def draw(self,surface):
    surface.blit(theme.surface,self.rect,self.theme_rect)
    self.scrolling_text.rect.topleft = (self.rect.left + self.text_x, self.rect.top + self.text_y)
    self.scrolling_text.draw(surface)
    if self.scrolling_subtext:
      self.scrolling_subtext.rect.topleft = (self.rect.left + self.text_x, self.rect.top + self.text_y+30)
      self.scrolling_subtext.draw(surface)

  def __cmp__(self,other):
    if not other: return 0
    return cmp(self.name.lower(),other.name.lower())

  def set_selected(self,selected):
    if selected != self.selected: # ignore no change
      self.selected = selected
      self.get_theme_rect()

  def set_focus(self,focus):
    self.scrolling_text.set_focus(focus)
    if self.scrolling_subtext: self.scrolling_subtext.set_focus(focus)
    print "set_focus: %s for %s" % (focus,self.name)

  def update(self):
    self.scrolling_text.update()
    if self.scrolling_subtext: self.scrolling_subtext.update()


class Song(Text):
  def __init__(self,track,title,path):
    self.track    = self._fix_track(track)
    self.title    = title
    self.path     = path
    self.track_font_surf= globals.gprint_ret(self.track)
    Text.__init__(self,self.title) # call Sprite initializer

  def _fix_track(self,track):
    if not track: track = ''
    else: track = "%02d " % (track,)
    return track

  def get_scrolling_text(self):
    if not self.scrolling_text:
      self.scrolling_text = ScrollingText(
          self.name,
          Rect(
            self.text_x + self.track_font_surf.get_rect().width,
            self.text_y,
            self.rect.width - (self.text_x * 2) - self.track_font_surf.get_rect().width,
            self.rect.height))

  def draw(self,surface):
    surface.blit(theme.surface,self.rect,self.theme_rect)
    surface.blit(self.track_font_surf,(self.rect.left + self.text_x,self.rect.top + self.text_y))
    self.scrolling_text.rect.topleft = (self.rect.left + self.text_x + self.track_font_surf.get_rect().width, self.rect.top + self.text_y)
    self.scrolling_text.draw(surface)
    if self.scrolling_subtext: self.scrolling_subtext.draw(surface)

  def __cmp__(self,other):
    if not other: return 0
    return cmp(self.track + self.name.lower(),other.track + other.name.lower())


# This is **not** a widget. It's more of a blit helper. It makes it easy to implement progress bar within a rectangle.
class ProgressBar():
  def __init__(self, rect, percent=0, color=None, segmented=True):
    ''' Displays a graphical bar representing a positive integer 
    between 0 and 100, inclusive '''
    self.rect      = rect
    self.percent   = int(percent)
    self.segmented = segmented
    if color == None: self.bar_color = theme.bar_color_sel
    else: self.bar_color = color

  def blit(self,image):
    max_bar_width = self.rect.right - self.rect.left
    current_bar_width = max_bar_width * self.percent / 100
    bar_segment_width = max_bar_width / 10
    if not self.segmented: image.fill(self.bar_color,Rect(self.rect.left,self.rect.top,current_bar_width,self.rect.bottom - self.rect.top))
    else:
      c = 0
      while c < current_bar_width:
        if c + bar_segment_width < current_bar_width:
          current_segment_width = bar_segment_width
        else:
          current_segment_width = current_bar_width - c
        image.fill(self.bar_color,Rect(self.rect.left + c,self.rect.top,current_segment_width - 1,self.rect.bottom - self.rect.top))
        c += bar_segment_width
      
  def set_color(self,color):
    self.bar_color = color

  def set_percent(self,percent):
    self.percent = int(percent)


class PercentBar(BaseWidget):
  def __init__(self,percent, selected=False):
    ''' Displays a positive integer between 0 and 100, inclusive
    along with a graphical bar representing this number '''
    BaseWidget.__init__(self)
    self.pbar      = None
    self.rect      = None
    self.percent   = 0
    self.font_surf = None
    self.selected  = selected
    self.set_percent(percent)

  def _load_image(self):
    if self.selected:
      self.image = theme.get_image('text_bg_sel')
      bar_color  = theme.bar_color_sel
    else:
      self.image = theme.get_image('text_bg')
      bar_color  = theme.bar_color

    offset = self.font_surf.get_rect().width + 5 + 5
    self.image.blit(self.font_surf,(5,33))
    if not self.rect: self.rect = self.image.get_rect()
    if not self.pbar: self.pbar = ProgressBar(
        Rect(offset,33,self.rect.right - self.rect.left - offset - 5,30),
        self.percent,
        bar_color)
    else:
      self.pbar.set_color(bar_color)
      self.pbar.set_percent(self.percent)
    self.pbar.blit(self.image)
      
  def __cmp__(self,other):
    if not other: return 0
    return cmp(self.percent,other.percent)

  def set_selected(self,selected):
    if selected != self.selected: # ignore no change
      self.selected = selected
      self._load_image()

  def set_focus(self,focus):
    ''' do nothing '''

  def set_percent(self,percent):
    self.percent = int(percent)
    self.font_surf = globals.gprint_ret("%3d%%" % (self.percent,))
    self._load_image()


class Bar(BaseWidget):
  def __init__(self,rect):
    BaseWidget.__init__(self)
    self.rect  = Rect(rect)
    self.image = theme.get_surface((self.rect.width,self.rect.height))
